unit KeyU;

{$A-} {pack records here}

{====================== KeyU MODULE ==========================

  This module contains the routines that handle all the TKey routines.
  The null string is not a valid key. But a blank is OK.

===============================================================}

interface

uses StdCtrls, ExtCtrls, Graphics, Forms, Buttons, Classes,
   Controls, Windows, Dialogs, SysUtils;

{$A-}  {Make sure these records are packed}

const
  NumLocs = 193;
  Offset = 145;

type
  TRecsPerKey = (None, One, Many);

  {See below for the use of each of these fields.}
  KeyHeader = record
    NextID:           longint;
    Active:           longint;
    KL:               word;
    RecsPerKey:       TRecsPerKey;
    RootLevel:        byte;
    SendNumber:       byte;      {When splitting a sector}
    KeepNumber:       byte;
    RootSector:       word;
    FirstSector:      word;
    end;
  KeySector = record
    KH:            KeyHeader;
    Level:         byte;
    Count:         byte;
    Follower:      word;
    Previous:      word;
    C:             array [0..2023] of byte;
    end;


  SectorFile = file of KeySector;
  SectorInfo = record
    SecNo:            word;
    ItemNo:           byte;
    end;
  IDSectorInfo = record
    SecNo:            word;
    ItemNo:           byte;
    end;
  SectorInfoArray = array [0..9] of SectorInfo;

  KeyData = record
    kjd:  longint;  {key id}
    rjd:  longint;  {rec id or start of rec id's}
    num: word;     {number of record id's}
    end;

  Locator = record
    at:   longint;
    num:  word;
    end;


type
  TKey = class (TComponent)
  private
    NextID:           longint;
    Active:           longint;
    KL:               word;
    FRecsPerKey:      TRecsPerKey;
    RootLevel:        byte;
    SendNumber:       byte;      {When splitting a sector}
    KeepNumber:       byte;
    RootSector:       word;
    FirstSector:      word;
    FFileName:        string;
    FKeyAttached:     boolean;
    procedure SetFileName(Value: string);
    procedure SetKeyLength(Value: word);
    procedure SetRecsPerKey(Value: TRecsPerKey);
  protected
    PKL:              byte;  {make visible to TKey and TMultiKey}
    KS:               KeySector;
    FreeList:         array [1..NumLocs] of Locator;
    FreeCount:        word;
    SIA:              SectorInfoArray;
    SecFile:          SectorFile;
    KeyIDFile:        file of word;
    RecIDFile:        file of longint;
    procedure ReadSector(Sector: word; var S: KeySector);
    procedure WriteSector(Sector: word; var S: KeySector);
    procedure PutInKey(var S: KeySector; k: shortstring;
                KD: KeyData; item: byte);
    procedure GetTheKey(var S: KeySector; var k: shortstring;
                 var KD: KeyData; item: byte);
    procedure FindItem(k: shortstring; var item: byte;
                 var f: shortstring; var KD: KeyData);
    procedure PrimativeFindKey(ByPass: boolean; k: shortstring;
                 var f: shortstring; var KD: KeyData);
    procedure WriteZero;
    procedure AddToFreeList(xat: longint; xnum: word);
    function  FindRoom(needed:word): longint;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure AttachFile;
    procedure DetachFile;
    function  ExistKey(sought: shortstring): boolean;
    function  FindKey(sought: shortstring): shortstring;
    function  FirstKey: shortstring;
    function  LastKey: shortstring;
    function  NextKey(key: shortstring): shortstring;
    function  PrevKey(key: shortstring): shortstring;
    function  InsertKey(newkey: shortstring): boolean;
    function  DeleteKey(killkey: shortstring; IfEmpty: boolean): boolean;
    function  ChangeKey(oldkey, newkey: shortstring): boolean;
    function  GetKeyID(key: shortstring): longint;
    function  GetKey(kid: longint): shortstring;
    {These routines deal with record ids.}
    function  GetRecIDCount(key: shortstring): word;
    function  IncludeRecID(key: shortstring; rid: longint): boolean;
    function  RemoveRecID(key: shortstring; rid: longint): boolean;
    function  GetRecID(key: shortstring; n: word): longint;
    {The following two routines are used by SelectU. Do not use.}
    procedure KeyByRatio(ratio: single; var found: shortstring);
    function  FindRatio(sought: shortstring): single;
  published
    property ActiveKeys: longint
      read  Active;
    property FileName: string
      read  FFileName
      write SetFileName;
    property KeyLength: word
      read  KL
      write SetKeyLength;
    property RecsPerKey: TRecsPerKey
      read  FRecsPerKey
      write SetRecsPerKey;
    end;

{A description of the properties and methods follows.}

{ActiveKeys:
 The number of keys in the file at the present time. Read only.

 FileName:
 The name of the disk file holding the keys.

 KeyLength:
 The maximum length of a key string. Read only but visible.

 RecsPerKey:
 Can have one of three states:
 None: File not attached yet. No key operations while in this state.
 One: One record id per key. Use this when the key string can be used to
 uniquely identify the record with the given id.
 Many: Many records per key. Use this when there can be many records
 involving the given key string. This property is set when the file
 is created and canoot be changed after that.


 AttachFile:
 This method MUST be called before any other methods can be executed.
 This method hooks the key data set with disk name FileName
 to the TKey.

 DetachFile:
 This method needs to be called only if you wish to use a TKey
 component with another key data set after using it with another one.
 For example suppose the program is running with TKey attached to
 'FIRST.KEY'. To switch to 'SECOND.KEY' do this

  AKey.DetachFile;
  AKey.FileName := 'SECOND.KEY';
  AKey.AttachFile;


 Create:
 Obvious;

 ExistKey:
 Returns TRUE if the given key exists and FALSE if it does not.

 FindKey:
 This routine looks for the key sought in the file. If it is
 present it just returns the string sought. If the key sought is not
 in the file but some next key after sought is in the file then it
 returns that key. If there is no sought key or following key it
 returns the null string.


 FirstKey:
 Returns the lowest or first key in the file or the null string
 if there are no keys.


 LastKey:
 Returns the highest or last key in the file or the null string
 if there are no keys.


 NextKey:
 Returns the key in the file next after the given key or the
 null sting if there is none.

 To go through all the keys code something like this.

    CurrentKey := ThisKey.FirstKey;
    while CurrentKey <> '' do begin
      ... do things with the key record, such as fetch a record ...
      CurrentKey := ThisKey.NextKey(CurrentKey);
      end;
    end;


 PrevKey:
 Returns the key in the file just before the given key or the
 null string if there is none.

 InsertKey:
 Attempts to insert the key with string newkey into the file. If the key
 newkey is not already in the file it will insert it and return TRUE.
 You can get the KeyID associated with the newkey by calling GetKeyID
 with newkey as parameter. If newkey is already in the file this method
 returns FALSE and does nothing else.


 DeleteKey:
 Attemps to delete the key with string killkey from the file. First the
 methods looks to see if the key killkey exists. If it does not then it
 returns FALSE and does nothing else. Assuming killkey exists the method
 next looks at the parameter IfEmpty. If FALSE the method deletes the key
 killkey and returns TRUE. If IfEmpty is TRUE then the key killkey will
 be deleted only if there are no associated record id's in which case
 TRUE is returned. If there are associated record id's then FALSE is
 returned and nothing else happens. In all cases when TRUE is returned
 the key killkey was deleted and if FALSE was returned it was not.

 ChangeKey:
 This method changes the key string from its present value, oldkey, to
 a new value, newkey. Any associated id(s) are preserved. If either
 oldkey does not already exist or newkey already does exist the routine
 returns FALSE and the change does not occur. Otherwise the routine
 returns TRUE and the change does occur.

 GetKeyID:
 This returns the key id associated with the string key. This longint
 can be stored in place of the much longer string. If the key does not
 exist, then -1 is returned to indicate -No such key-.

 GetKey:
 This returns the key string associated with the given key id kid. If
 kid does not refer to a valid key then the null string is returned.

 Get RecIDCount:
 This returns the number of record id's assigned to the given key. If
 the RecsPerKey property has the value One, this value can only be
 0 or 1.

 IncludeRecID:
 This attempts to add the given record id rid to the key with string
 key. The method first sees if the key exists. If it does not, it returns
 FALSE and nothing else happens. Assuming that the key exists, the
 method next looks at the RecsPerKey property. If this property has the
 value One then only one record id is permitted. If RecsPerKey is One
 and there is already an assigned record id, the method returns FALSE
 and nothing else happens. If RecsPerKey is One and no record id is
 currently assigned, then rid is assigned and TRUE is returned. If the
 key exists AND RecsPerKey is Many then rid will be included among the
 assigned id's whether or not is was already present. In this case
 TRUE is returned. In general, if TRUE was returned then the given key
 exists and the given rid is among its assigned record id's.

 RemoveRecID:
 Attempts to remove the record id rid from the assigned id's for the
 key with the given string. If the key does not exist, FALSE is returned
 and nothing else happens. Assuming that the key exists then the given
 rid will be removed from the assigned record id's whether is was there
 or not and TRUE is returned. In any case, when TRUE is returned, the
 key exists and rid is not part of its assigned record id's.

 GetRecID:
 This method returns the record id in position n of the assigned record
 id's for the key with string key. The parameter n is 1 based (unlike most
 things in Delphi) so that to get the first (or only) id use n=1. If the
 record id does not exist for any reason, then -1 is returned. This can
 happen if the key does not exist, or it does but the index n is greater
 or equal to the number of assigned record ids. If the RecsPerKey property
 is One, the n parameter is ignored and the returned longint is either
 -1 or the single record id.}



procedure Register;


implementation

{ $R KEYU.RES}

{Here we describe how the TKey file is built and how it works.

There are two files: SecFile and KeyIDFile if RecsPerKey has the
value One and three files: adding RecIDFile if it has the value Many.

Sector 0 of the SecFile contains the 18 byte Key Header data
described shortly. All other sectors allow for these 18 bytes but
they contain no useful information there. Then there are 6 more
bytes that are unique to that sector. They tell how much information
it contains and how it is linked to other sectors.

The remaining information consists of a series of key/data pairs.
The key is a string (all of the same maximun size) and data is a
10 byte structure of type KeyData. KeyData is a record with three
fields. The first is kjd, a longint, that can be used in database
records as a shorthand for the longer key. Hence a string can appear
in many places of a database and use only 4 bytes for each occurance.
TKey has methods that get this longint, or given a longint get the
string. The next field is rjd, a longint, whose use depends on the
value of the RecsPerKey property. If RecsPerKey has the value One
then rjd is either -1 indicating -no assigned id- or the record id
of an assigned record. This is intended for situations in which the
string is the primary key to a record. If RecsPerKey has the value
Many then rjd is either -1 indicating -no assigned id- or the position
in the RecIDFile where the record id's assigned to this key are located.
The final field is num, a word, whose value is the number of record
id's assigned to this key. If RecsPerKey has the value One, then num
can only have the values 0 or 1. If, for example, some key/data pair
has rjd = 297 and num = 5 it means that the record id's assigned to
this key can be found in positions 297, 298, 299, 300, 301 in the
RecIDFile.

The Count field od a sector contains the number of key/data pairs
in the sector. When a key/data pair is inserted or deleted, this Count
is adjusted. The total number of key/data pairs in a sector possible
is the sum of the KeepNumber and SendNumber. When a sector is filled
with keys it is 'split' -- a new sector is created and half the keys
(SendNumber) sent to the new sector. Each key sector has pointers to
the previous sector and the following sector (in alphabetic order of
the keys) and these and the Count are adjusted when a sector is split.
What happen next depends on the 'level' of the sector.

The data part of the key/data in a level 0 sector is as described
above. However if the level is higher than 0 then the num part is
instead the sector number of a next lower level sector that will
contains the key part of the key/data as its last entry. Hence the
file has a tree form with the leaves at level 0 and leaves at
higher levels pointers to lower level limbs. When FindKey runs
it starts at the single sector at the highest level -- the root
sector. (The RootLevel and RootSector fields keep track of this
frequently consulted sector.). It finds the key or the next higher
and gets a pointer to the next lower level sector to search. This
procedure is repeated recursively until level 0 is reached when
the process stops.

Because of this tree structure whenever a sector is split, an entry
must be inserted in the next higher level to point to a new sector.
This may cause that sector to split and require yet another sector
to be created and a new entry inserted in yet a higher level. And
so on. In some cases it might even cause the root sector to split
so that the RootLevel and RootSector fields need to be changed.
When sectors are split the upper part of the sector (lower in
alphabetic order) are moved to the new sector and the  lower part
kept in the old sector. Hence sector 0 will IfEmpty have the highest
keys and the sector with the lowest keys could be anywhere. The field
FirstSector keeps track of this sector neded by the StartKey method.

Notice that in all cases the key part of the key/data pair in any
level higher than 0 is the highest key in the next lower sector
pointed to. When deletes are made, it could be this last key.
Rather than change the next higher level key/data we just set
the num part of the key/data in the level 0 sector to 65535 indicating
this condition. The TKey methods are aware of this possiblilty and
behave correctly by bypassing these 'hidden' keys. Hence deletes
affect only level 0 sectors -- a great savings of time and code.

When inserting a new key/data pair the method uses the SIA array.
Inserts start by calling FindKey to set the SIA array. As the
tree is scanned from highest level to lowest, the sectors
encountered and the items numbers of the found keys are recorded
in the SIA array. This data is used to split sectors if needed.
Also when a new key in inserted a new keyid is created for it and
stored along with the key string and its kjd field. The rjd and num
fields of new keys are set to -1 and 0 respectively.


Here is a brief descriptions of the fields in a KeyHeader.

  NextID: The keyid for the next key to be inserted.
  Active: The number of keys in the file.
  KL: byte;  The maximum length of a key. This is set when
      the file is created.
  FRecsPerKey: Set to None then the TKey is created, and
      changed to One or Many when the KeyFile is attached.
  RootLevel: byte:  The level of the root in the tree. There are
      statements of the form
        ... for level := KeyRootLevel downto 0 do ...
  RootSector: word;  The sector number of the root key sector --
      originally 0.
  FirstSector: word;  The sector number of the key sector with the
      lowest key -- originally 0.
  SendNumber: byte; The number of key/loc pairs sent to a new
      sector when a sector is split. Set when the key file is
      created.
  KeepNumber: byte; The number of key/loc pairs kept in the old
      sectorwhen it is split. Set when the key file is created.
      The sum of SendNumber and KeepNumber is the maximum number
      sector.


Here is a brief description of the fields in the common part of sectors.

  Level: byte; The level in the tree of this sector;
  Count: byte; The number of key/loc or id/sec pairs in the sector;
  Follower: word; The sector number of the sector immediately
      following this one in order. 65535 marks the end.
  Previous: word; The sector number of the sector immediately
      preceding this one. 65535 marks the beginning.


The KeyIDFile is just a sequence of words. Each word represents a sector
number of the SecFile which contains the key/data pair having the
kjd field equal to the position of the word in the KeyIDFile. If
the word is position 459 is 23 it means that sector 23 contains the
key/data pair with kjd = 459. This sector can be quickly searched
to find the key string. This file allows keyids and key strings to
be easily converted back and forth.

The RecIDFile has two parts: a 1160 byte prefix set up as an array
of 193 LocItems and a word, then a sequence of longints that represent
record id's. The array of LocItems is the 'FreeList'. If the record id's
assigned to a gien key change the place in the RecIDFile holding that
list of id's may change. The old place becomes 'free' and could be reused
by other keys. The FreeList keeps track of all this. }


type
  IDArray = array [1..16384] of longint;
  IDArrayPtr = ^IDArray;


{These procedures are useful but not directly called.}

Procedure TKey.ReadSector(Sector: word; var S: KeySector);
{Reads sector Sector into S from KeyFile.}

begin
  Seek(SecFile,Sector);
  Read(SecFile,S);
  end;


Procedure TKey.WriteSector(Sector: word; var S: KeySector);
{Writes Sector S into KeyFile at position Sector.}

begin
  Seek(SecFile,Sector);
  Write(SecFile,S);
  end;


Procedure TKey.PutInKey(var S: KeySector; k: shortstring;
              KD: KeyData; item: byte);
{Put the key and KeyData into the KeySector S at place item. }

begin
  move(k,S.C[(KL+11)*pred(item)],succ(KL));
  move(KD,S.C[(KL+11)*item-10],10);
  end;


Procedure TKey.GetTheKey(var S: KeySector; var k: shortstring;
            var KD: KeyData; item: byte);
{Returns the key and KeyData from KeySector S from place item.}

begin
  move(S.C[(KL+11)*pred(item)],k,succ(KL));
  move(S.C[(KL+11)*item-10],KD,10);
  end;


Procedure TKey.FindItem(k: shortstring; var item: byte;
              var f: shortstring; var KD: KeyData);
{This procedure finds the position of the sought key k in the
 KeySector KS.  In all cases the sought key k is <= the
 last key in the sector. It returns the item number in item,
 the found key in f.  Item is the slot containing the sought
 key if k = found and the place where the sought key k should
 be inserted if k <> found.}

var
  last, index: byte;
  done: boolean;

begin
  item := 1; last := KS.Count; done := FALSE;
  {binary search is used.}
  while (item <= last) and not done do begin
    index := (item + last) div 2;
    GetTheKey(KS,f,KD,index);
    if k = f
    then begin
      done := TRUE;
      item := index;
      end
    else if k < f
         then last := index - 1
         else item := index + 1;
    end;
  if not done
  then GetTheKey(KS,f,KD,item);
  end;



Procedure TKey.PrimativeFindKey(ByPass: boolean; k: shortstring;
              var f: shortstring; var KD: KeyData);
{Finds the key. k is the sought key, f is the found key. As
a side effect the routine sets up data in the SIA array of KI.
This data is used during inserts and deletes. If ByPass is TRUE
then the user is doing a FindKey and any final deleted keys
in a sector are bypassed. If ByPass is FALSE the user is doing
an insert and a final deleted key in a sector is returned as
a regular result.}

var
  SeekSector: longint;
  item,lev: byte;

begin
  SeekSector := RootSector;
  for lev := RootLevel downto 0 do begin
    ReadSector(SeekSector,KS);
    FindItem(k,item,f,KD);
    SIA[lev].SecNo := SeekSector;
    SIA[lev].ItemNo := item;
    SeekSector := KD.num;
    end;
  if ByPass and (KD.num = 65535) then with KS do
    while (SIA[0].ItemNo = Count) and (KD.num = 65535)
    do begin
      {found a final deleted item}
      SeekSector := Follower;
      ReadSector(SeekSector,KS);
      GetTheKey(KS,f,KD,1);
      SIA[0].SecNo := SeekSector;
      SIA[0].ItemNo := 1;
      end;
  end;



procedure TKey.AddToFreeList(xat: longint; xnum: word);

var
  i: word;

begin
  if FreeCount = NumLocs
  then exit; {no room for another entry}
  if FreeCount = 0
  then begin
    FreeList[1].at := xat;
    FreeList[1].num := xnum;
    inc(FreeCount);
    exit;
    end;
  i := 1;
  while (i <= FreeCount) and (FreeList[i].at < xat) do inc(i);
  if i <= FreeCount
  then move(FreeList[i],FreeList[i+1],(FreeCount-i+1)*6);
  FreeList[i].at := xat;
  FreeList[i].num := xnum;
  inc(FreeCount);
  {Now collapse if possible}
  if i = 1
  then i := 2;
  while (FreeList[i-1].at+FreeList[i-1].num = FreeList[i].at)
  do begin
    FreeList[i-1].num := FreeList[i-1].num+FreeList[i].num;
    if i+1<=FreeCount
    then move(FreeList[i+1],FreeList[i],(FreeCount-i)*6);
    dec(FreeCount);
    end;
  end;




function TKey.FindRoom(needed: word): longint;

var
  i: word;

begin
  Result := FileSize(RecIDFile);
  i := 1;
  while i <= FreeCount do
    if FreeList[i].num = needed
    then begin
      Result := FreeList[i].at;
      if i < FreeCount
      then move(FreeLIst[i+1],FreeList[i],6*(FreeCount-i));
      dec(FreeCount);
      exit;
      end
    else if FreeList[i].num > needed
         then begin
           Result := FreeList[i].at;
           FreeList[i].at := FreeList[i].at + needed;
           FreeList[i].num := FreeList[i].num - needed;
           exit;
           end
         else inc(i);
  end;



procedure TKey.WriteZero;

var
  S: KeySector;

begin
  seek(SecFile,0);
  read(SecFile,S);
  move(NextID,S,sizeof(KeyHeader));
  seek(SecFile,0);
  write(SecFile,S);
  end;


{----------------------------------------------------------------}
{                                                                }
{ Methods of Key                                                 }
{                                                                }
{----------------------------------------------------------------}

{================================================================}
{     Routines that set properties                               }
{================================================================}

procedure TKey.SetFileName(Value: string);
{Note that the file is opened to read KeyLength and RecsPerKey.
If RecsPerKey is Many the RecIDFile is also opened. If the file is
not there the routine complains.}

var
  SecFile: file of KeySector;
  KeyIDFile: file of word;
  RecIDFile: file of longint;
  S: KeySector;

begin
  if FFileName <> Value
  then begin
    if Value = ''
    then if FKeyAttached
         then DetachFile
         else
    else begin
      AssignFile(SecFile,Value);
      {$I-} reset(SecFile); {$I+}
      if IOResult <> 0
      then begin
        ShowMessage('The file '+Value+' does not exist');
        FFileName := '';
        exit;
        end;
      seek(SecFile,0);
      Read(SecFile,S);
      KL := S.KH.KL;
      FRecsPerKey := S.KH.RecsPerKey;
      CloseFile(SecFile);

      AssignFile(KeyIDFile,ExtractFileDir(Value)+
                    '_'+ExtractFileName(Value));
      {$I-} reset(KeyIDFile); {$I-}
      if IOResult <> 0
      then begin
        ShowMessage('The file '+ExtractFileDir(Value)+
           '_'+ExtractFileName(Value)+' does not exist.');
        FFileName := '';
        exit;
        end;
      CloseFile(KeyIDFile);

      if RecsPerKey = Many
      then begin
        AssignFile(RecIDFile,ExtractFileDir(Value)+
                      '_X'+ExtractFileName(Value));
        {$I-} reset(RecIDFile); {$I-}
        if IOResult <> 0
        then begin
          ShowMessage('The file '+ExtractFileDir(Value)+
             '_X'+ExtractFileName(Value)+' does not exist.');
          FFileName := '';
          exit;
          end;
        CloseFile(RecIDFile);
        end;
      end;;
    FFileName := Value;
    end;
  end;


{The following null procedure make the KeyLength  visible in the
Object Inspector. Of course it is not to be changed.}

procedure TKey.SetKeyLength(Value: word);

begin
  end;


procedure TKey.SetRecsPerKey(Value: TRecsPerKey);

begin
  end;


{==========================================================}
{    User interface methods                                }
{==========================================================}

constructor TKey.Create(AOwner: TComponent);

begin
  inherited Create(AOwner);
  FFileName := '';
  FKeyAttached := FALSE;
  end;


destructor TKey.Destroy;

begin
  DetachFile;
  inherited Destroy;
  end;



procedure TKey.AttachFile;

begin
  if FKeyAttached
  then begin
    ShowMessage('File already attached');
    exit;
    end;
  FKeyAttached := FALSE; {Assume the worst}
  if FFileName = ''
  then begin
    ShowMessage('No file named yet.');
    exit;
    end;
  AssignFile(SecFile,FFileName);
  reset(SecFile);
  seek(SecFile,0);
  Read(SecFile,KS);
  Move(KS,NextID,sizeof(KeyHeader));
  AssignFile(KeyIDFile,ExtractFileDir(FFileName)+'_'+
        ExtractFileName(FFileName));
  reset(KeyIDFile);
  if FRecsPerKey = Many
  then begin
    AssignFile(RecIDFile,ExtractFileDir(FFileName)+'_X'+
          ExtractFileName(FFileName));
    reset(RecIDFile);
    BlockRead(RecIDFile,FreeList,Offset);
    end;
  FKeyAttached := TRUE;
  end;


procedure TKey.DetachFile;

begin
  If not FKeyAttached
  then exit;
  if FFileName <> ''
  then begin
    CloseFile(SecFile);
    CloseFile(KeyIDFile);
    if RecsPerKey = Many
    then begin
      seek(RecIDFile,0);
      BlockWrite(RecIDFile,FreeList,Offset);
      CloseFile(RecIDFile);
      end;
    end;
  FFileName := '';
  FKeyAttached := FALSE;
  end;


function TKey.FindKey(sought: shortstring): shortstring;

var
  found: shortstring;
  KD: KeyData;

begin
  if ActiveKeys = 0
  then begin
    Result := '';
    exit;
    end;
  if sought = ''
  then sought := ' ';
  PrimativeFindKey(TRUE,sought,found,KD);
  {if found is the sentinal at the end of file its first
   character is chr(255).}
  if found[1] = chr(255)
  then Result := ''
  else Result := found;
  end;


function TKey.ExistKey(sought: shortstring): boolean;

var
  f: shortstring;
  KD: KeyData;

begin
  if ActiveKeys = 0
  then begin
    Result := FALSE;
    exit;
    end;
  PrimativeFindKey(TRUE,sought,f,KD);
  Result := (f[1] <> chr(255)) and (sought = f);
  end;



function TKey.FirstKey: shortstring;

var
  found: shortstring;
  KD: KeyData;

begin
  if ActiveKeys = 0
  then begin
    Result := '';
    exit;
    end;
  ReadSector(FirstSector,KS);
  GetTheKey(KS,found,KD,1);
  if KD.kjd <> -1
  then Result := found
  else Result := '';
  end;


function TKey.LastKey: shortstring;

var
  found: shortstring;
  KD: KeyData;

begin
  if ActiveKeys = 0
  then begin
   Result := '';
   exit;
   end;
  ReadSector(0,KS);
  if KS.Count = 1
  then Result := ''
  else begin
    GetTheKey(KS,found,KD,KS.Count-1);
    Result := found;
    end;
  end;


function TKey.NextKey(key: shortstring): shortstring;

var
  found: shortstring;
  KD: KeyData;

Procedure NextSector;

begin
  SIA[0].SecNo := KS.Follower;
  ReadSector(KS.Follower,KS);
  SIA[0].ItemNo := 1;
  end;

begin
  if ActiveKeys = 0
  then begin
    Result := '';
    exit;
    end;
  PrimativeFindKey(TRUE,key,found,KD);
  with KS do begin
    if (SIA[0].ItemNo = Count-1) and (SIA[0].SecNo = 0)
    then begin
      Result := '';
      exit;
      end;
    inc(SIA[0].ItemNo);
    if SIA[0].ItemNo > Count
    then NextSector;
    GetTheKey(KS,found,KD,SIA[0].ItemNo);
    if KD.num = 65535
    then while (SIA[0].ItemNo = Count) and (KD.num = 65535)
      do begin
        {found a final deleted item}
        NextSector;
        GetTheKey(KS,found,KD,SIA[0].ItemNo);
        end;
    if found[1] = chr(255)
    then Result := ''
    else Result := found;
    end;
  end;


function TKey.PrevKey(key: shortstring): shortstring;

var
  found: shortstring;
  KD: KeyData;

Procedure PrevSector;

begin
  SIA[0].SecNo := KS.Previous;
  ReadSector(KS.Previous,KS);
  SIA[0].ItemNo := KS.Count;
  end;

begin
  if ActiveKeys = 0
  then begin
    Result := '';
    exit;
    end;
  PrimativeFindKey(TRUE,key,found,KD);
  with KS do begin
    if (SIA[0].ItemNo = 1) and (SIA[0].SecNo = FirstSector)
    then begin
      Result := '';
      exit;
      end;
    dec(SIA[0].ItemNo);
    if SIA[0].ItemNo = 0
    then PrevSector;
    GetTheKey(KS,found,KD,SIA[0].ItemNo);
    if KD.num = 65535
    then while (SIA[0].ItemNo = Count) and (KD.num = 65535)
      do begin
        {found a final deleted item}
        if SIA[0].ItemNo = 1
        then PrevSector
        else dec(SIA[0].ItemNo);
        GetTheKey(KS,found,KD,SIA[0].ItemNo);
        end;
    if found[1] = chr(255)
    then Result := ''
    else Result := found;
    end;
  end;



function TKey.InsertKey(newkey: shortstring): boolean;

var
  f: shortstring;
  KD,Split: KeyData;
  NewSec,A,B,Z,w: word;
  lev: byte;
  done: boolean;
  S: KeySector;

begin
  Result := TRUE;
  if newkey = ''
  then begin {Trying to add the null string}
    Result := FALSE;
    exit;
    end;

  PrimativeFindKey(FALSE,newkey,f,KD);
  if (newkey = f) and (KD.num <> 65535)
  then begin {Trying to add a duplicate key}
    Result := FALSE;
    exit;
    end;

  done := FALSE; lev := 0;
  KD.kjd := NextID; KD.rjd := -1; KD.num := 0;
  while not done do begin
    with KS do begin
      {make room for new item}
      ReadSector(SIA[lev].SecNo,KS);
      Move(C[pred(SIA[lev].ItemNo)*(KL+11)],     {Move keys}
           C[SIA[lev].ItemNo*(KL+11)],
           (Count-SIA[lev].ItemNo+1)*(KL+11));
      {insert new item}
      PutInKey(KS,newkey,KD,SIA[lev].ItemNo);
      inc(Count);
      WriteSector(SIA[lev].SecNo,KS);
      if lev = 0
      then begin
        seek(KeyIDFile,NextID);
        write(KeyIDFile,SIA[0].SecNo);
        end;

      if Count < SendNumber + KeepNumber
      then done := true
      else begin  {split sector}
        NewSec := FileSize(SecFile);
        {keep the mid item safe}
        GetTheKey(KS,newkey,KD,SendNumber);
        KD.num := NewSec;
        A := SIA[lev].SecNo; B := Follower; Z := Previous;
        Count := SendNumber;
        Follower := A;
        Previous := Z;
        WriteSector(NewSec,KS);
        if lev = 0
        then for w := 1 to Count do begin
          GetTheKey(KS,f,Split,w);
          seek(KeyIDFile,Split.kjd);
          write(KeyIDFile,NewSec);
          end;

        Move(C[(SendNumber)*(KL+11)],
             C,KeepNumber*(KL+11));
        Count := KeepNumber;
        Follower := B;
        Previous := NewSec;
        WriteSector(A,KS);
        if Z = 65535
        then if lev = 0
             then FirstSector := NewSec
             else
        else begin
          ReadSector(Z,S);
          S.Follower := NewSec;
          WriteSector(Z,S);
          end;
       inc(lev);

       if lev > RootLevel
       then begin  {new root needed}
         PutInKey(KS,newkey,KD,1);
         fillchar(f,sizeof(f),255);
         SetLength(f,KL);
         KD.num := SIA[lev-1].SecNo;
         PutInKey(KS,f,KD,2);
         Level := lev;
         Count := 2;
         Follower := 65535;
         Previous := 65535;
         RootSector := FileSize(SecFile);
         WriteSector(RootSector,KS);
         RootLevel := lev;
         done := TRUE;
         end;
       end;
      end;
    end;
  inc(Active);
  inc(NextID);
  WriteZero;
  end;



function TKey.DeleteKey(killkey: shortstring; IfEmpty: boolean):
               boolean;

var
  f: shortstring;
  KD: KeyData;

begin
  if ActiveKeys = 0
  then begin
    Result := FALSE;
    exit;
    end;
  PrimativeFindKey(TRUE,killkey,f,KD);
  if (f <> killkey) or (IfEmpty and (KD.num <> 0))
  then begin
    Result := FALSE;
    exit;  {do nothing}
    end
  else Result := TRUE;

  with KS do begin
    if (FRecsPerKey = Many) and (KD.num > 0)
    then AddToFreeList(KD.rjd,KD.num);
    seek(KeyIDFile,KD.kjd);
    KD.num := 65535;
    write(KeyIDFile,KD.num);
    if SIA[0].ItemNo <> Count
    then begin
      {cover up deleted item}
      Move(C[SIA[0].ItemNo*(KL+11)],     {Move keys}
           C[pred(SIA[0].ItemNo)*(KL+11)],
           (Count-SIA[0].ItemNo)*(KL+11));
      dec(KS.Count);
      end
    {else just insert 65535 as num}
    else PutInKey(KS,killkey,KD,Count);
    WriteSector(SIA[0].SecNo,KS);
    dec(Active);
    WriteZero;
    end;
  end;


function TKey.ChangeKey(oldkey, newkey: shortstring): boolean;

var
  KD,Safe: KeyData;
  akey: shortstring;

begin
  if ActiveKeys = 0
  then begin
    Result := FALSE;
    exit;
    end;
  Result := TRUE; {assume the change will work}
  PrimativeFindKey(TRUE,newkey,akey,KD);
  if akey = newkey
  then begin  {newkey already exists, do not change}
    Result := FALSE;
    exit;
    end;
  PrimativeFindKey(TRUE,oldkey,akey,Safe);
  if akey <> oldkey
  then begin  {oldkey does not exist. do not change}
    Result := FALSE;
    exit;
    end;
  DeleteKey(oldkey,FALSE);
  InsertKey(newkey);
  dec(NextID); {Insert incremented NextID. Decrement to restore.}
  PrimativeFindKey(TRUE,newkey,akey,KD); {Set SIA array}
  PutInKey(KS,newkey,Safe,SIA[0].ItemNo);
  WriteSector(SIA[0].SecNo,KS);
  seek(KeyIDFile,Safe.kjd);
  Write(KeyIDFile,SIA[0].SecNo);
  WriteZero;
  end;



function TKey.GetKeyID(key: shortstring): longint;

var
  KD: KeyData;

begin
  if ExistKey(key)
  then begin
    GetTheKey(KS,key,KD,SIA[0].ItemNo);
    Result := KD.kjd;
    end
  else Result := -1;
  end;



function TKey.GetKey(kid: longint): shortstring;

var
  Sec,w: word;
  f: shortstring;
  KD: KeyData;

begin
  if (kid >= filesize(KeyIDFile)) or (ActiveKeys = 0)
  then Result := ''
  else begin
    seek(KeyIDFile,kid);
    read(KeyIDFile,Sec);
    if Sec = 65535
    then Result := ''
    else begin
      ReadSector(Sec,KS);
      w := 0;
      repeat
        inc(w);
        GetTheKey(KS,f,KD,w);
        until (KD.kjd = kid);
      Result := f;
      end;
    end;
  end;


function TKey.IncludeRecID(key: shortstring; rid: longint): boolean;

var
  found: shortstring;
  KD: KeyData;
  p: IDArrayPtr;
  w,size: word;

begin
  Result := ExistKey(key);
  if Result
  then begin
    GetTheKey(KS,found,KD,SIA[0].ItemNo);
    case FRecsPerKey of
      One: if KD.num = 0
           then begin {No assigned record id's yet.}
             KD.rjd := rid; KD.num := 1;
             PutInKey(KS,found,KD,SIA[0].ItemNo);
             WriteSector(SIA[0].SecNo,KS);
             end
           else  {There is an assigned id. It should be rid}
             if KD.rjd <> rid
             then Result := FALSE;
      Many: if KD.num = 0
            then begin {No assigned id's yet.}
              KD.rjd := FindRoom(1);
              KD.num := 1;
              PutInKey(KS,found,KD,SIA[0].ItemNo);
              WriteSector(SIA[0].SecNo,KS);
              seek(RecIDFile,KD.rjd);
              Write(RecIDFile,rid);
              end
            else begin
              size := 4*KD.num+4; {make room for one more.}
              GetMem(p,size);
              seek(RecIDFile,KD.rjd);
              BlockRead(RecIDFile,p^,KD.num);
              p^[KD.num+1] := rid;
              w := 1;
              while (p^[w] <> rid) do inc(w);
              if w > KD.num
              then begin
                AddToFreeList(KD.rjd,KD.num);
                KD.num := KD.num + 1;
                KD.rjd := FindRoom(KD.num);
                PutInKey(KS,found,KD,SIA[0].ItemNo);
                WriteSector(SIA[0].SecNo,KS);
                seek(RecIDFile,KD.rjd);
                BlockWrite(RecIDFile,p^,KD.num);
                end;
              FreeMem(p,size);
              end;
      end;
    end;
  end;


function TKey.RemoveRecID(key: shortstring; rid: longint): boolean;

var
  found: shortstring;
  KD: KeyData;
  p: IDArrayPtr;
  w,size: word;

begin
  Result := ExistKey(key);
  if Result
  then begin
    GetTheKey(KS,found,KD,SIA[0].ItemNo);
    case FRecsPerKey of
      One: if KD.rjd = rid
           then begin
             KD.rjd := -1; KD.num := 0;
             PutInKey(KS,found,KD,SIA[0].ItemNo);
             WriteSector(SIA[0].SecNo,KS);
             end;
      Many: if KD.num > 0
            then begin
              size := 4*KD.num+4;
              GetMem(p,size);
              seek(RecIDFile,KD.rjd);
              BlockRead(RecIDFile,p^,KD.num);
              p^[KD.num+1] := rid;
              w := 1;
              while p^[w] <> rid do inc(w);
              if w <= KD.num
              then begin
                AddToFreeList(KD.rjd,KD.num);
                KD.num := KD.num - 1;
                if KD.num > 0
                then KD.rjd := FindRoom(KD.num)
                else KD.rjd := -1;
                PutInKey(KS,found,KD,SIA[0].ItemNo);
                WriteSector(SIA[0].SecNo,KS);
                if KD.num > 0
                then begin
                  p^[w] := p^[KD.num + 1];
                  seek(RecIDFile,KD.rjd);
                  BlockWrite(RecIDFile,p^,KD.num);
                  end;
                end;
              FreeMem(p,size);
              end;
      end;
    end;
  end;



function TKey.GetRecIDCount(key: shortstring): word;

var
  KD: KeyData;
  akey: shortstring;

begin
  if ExistKey(key)
  then begin
    GetTheKey(KS,akey,KD,SIA[0].ItemNo);
    Result := KD.num;
    end
  else Result := 0;
  end;


function TKey.GetRecID(key: shortstring; n: word): longint;

var
  found: shortstring;
  KD: KeyData;

begin
  if ExistKey(key)
  then begin
    GetTheKey(KS,found,KD,SIA[0].ItemNo);
    if RecsPerKey = One
    then Result := KD.rjd
    else if (KD.num = 0) or (n > KD.num)
         then Result := -1
         else begin
           seek(RecIDFile,KD.rjd+n-1);
           Read(RecIDFile,Result);
           end
    end
  else Result := -1;
  end;


procedure TKey.KeyByRatio(ratio: single;
                  var found: shortstring);

var
  SeekSector: word;
  item: byte;
  wanted,have: longint;
  S: KeySector;
  KD: KeyData;

begin
  if ActiveKeys = 0
  then begin
    found := '';
    exit;
    end;
  {Adjust ratio to be in the interval [0,1]}
  if ratio > 0.99
  then begin
    found := LastKey;
    exit;
    end;
  if ratio < 0.01
  then begin
    found := FirstKey;
    exit;
    end;
  wanted := Round(ratio*Active); have := 0;
  SeekSector := FirstSector;
  while have < wanted do begin
    ReadSector(SeekSector,S);
    have := have + S.Count;
    SeekSector := S.Follower;
    end;
 item := have - wanted;
 GetTheKey(S,found,KD,item);
 end;




function TKey.FindRatio(sought: shortstring): single;

var
  found: shortstring;
  SeekSector: word;
  have: longint;
  S: KeySector;
  KD: KeyData;

begin
  if ActiveKeys = 0
  then begin
    Result := 0.0;
    exit;
    end;
  SeekSector := FirstSector;
  ReadSector(SeekSector,S);
  GetTheKey(S,found,KD,S.Count);
  have := S.Count;
  while found < sought do begin
    SeekSector := S.Follower;
    ReadSector(SeekSector,S);
    GetTheKey(S,found,KD,S.Count);
    have := have + S.Count;
    end;
  Result := have / Active;
  end;



procedure Register;
begin
  RegisterComponents('Len''s',[TKey]);
  end;


end.

