{This unit contains all the routines needed for handling files with variable
length (up to 65528 bytes) records.  Records are identified by positive
longints and all of the routines refer to these numbers. The numbers do
NOT indicate where the records are stored in the file. That information
is in a seperate file maintained by these routines.  To use any of the
routines just drop a TVFile component on a form, set its FileName
property and call AttachFile to use the routines. (See below for
details). Use the Create Wizard to create a VFile data set.}

unit VFileU;

{$A-}  {pack records here}


interface

uses StdCtrls, ExtCtrls, Graphics, Forms, Buttons, Classes,
   Controls, Windows, Dialogs, SysUtils;

{$A-}  {Make sure these records are packed}


type
  LocItem = record
    start:          longint;
    count:          longint;
    end;
  FreeList = array [0..0] of LocItem;
  FreeListPtr = ^FreeList;

type
  TVFile = class(TComponent)
  private
    RecCount:       longint;
    BlockSiz:       longint;
    FreeCapacity:   longint;
    FreeCount:      longint;
    MaxCount:       longint;
    FLP:            FreeListPtr;
    GenLI:          LocItem;
    LocF:           file;
    DataF:          file;
    FFileName:      string;
    FAttached:      boolean;
    procedure SetVFileName(Value: string);
    procedure SetBlockSize(Value: longint);
    function  FindArea(BlocksWanted: longint): longint;
    procedure AddToFreeList(GivenStart,GivenCount: longint);
 published
    constructor Create(AOwner: TComponent); override;
    destructor  Destroy; override;
    procedure AttachFile;
    procedure DetachFile;
    function  GetBufSize(rid: longint): longint;
    function  Exist(rid: longint): boolean;
    function  Fetch(rid: longint; Buf: pointer): boolean;
    function  Insert(rid: longint; Buf: pointer; RL: longint): boolean;
    function  Restore(rid: longint; Buf: pointer; RL: longint): boolean;
    function  Delete(rid: longint): boolean;
    property FileName: string
       read  FFileName
       write SetVFileName;
    property RecordCount: longint
       read  RecCount;
    property BlockSize: longint
       read  BlockSiz
       write SetBlockSize;
    end;

{FileName:  This is the name of a data set that contains the actual records.
There is also an auxiliary file with the same name prefixed with a underscore.
This file contains the locations of the records and is consulted when records
are fetched, inserted, deleted or restored. The data set must exixt on disk
when this property is set since SetVFileName reads it to set some internal
fields.

BlockSiz: This is the size of each block. Read only.

RecordCount:  Just the number of records in the file. Read only.

AttachFile: This method hooks up the internal TVFile with the disk data
set named in the FileName property. It MUST be called before any other
methods or they wont work. A good place to put this is in the OnCreate
event for the form.

DetachFile. This method need not be called unless you want to use the
TVFile on another data set. This method breaks the connection between
the data set and the TVFile. Here is an example. We wish to disconnect
the TVFile from the data set 'DATA1' and connect instead to 'DATA2'.

   AVFile.DetachFile;  -- disconnects DATA1 --
   AVFile.FileName := 'DATA2';
   AVFile.AttachFile;

GetBufSize: This method gets the size of a buffer (in bytes) that will
to hold the record with id rid.

Exist: Returns TRUE if the record with record number rid exists.

Fetch: This method gets the record with id rid and places it in the
buffer pointed to by Buf. The buffer must be large enough to hold the
record and this method will overwrite the buffer if it is not big
enough. FALSE is returned if the record does not exist.

Insert: This method is used to insert a new record with id rid. The
record is stored is the buffer pointer to by Buf. The user must supply
the record length in RL. This method checks to see that the id rid is
not being used already, and if it is, returns FALSE and does nothing.

Restore: This method is used to restore a fetched record. The id is rid.
The user must supply the record length since the record may have changed
size. This method checks to see that the id rid is being used already,
and if it is not, returns FALSE and does nothing.

Delete: This method is used to delete a record. The id is rid. This
method check to see that the id rid is being used already, and if not,
returns FALSE and does nothing.}


procedure Register;


implementation


procedure Register;
begin
  RegisterComponents('Len''s',[TVFile]);
  end;


{Here we describe how the TVFile is built and how it works.

The TVFile is actually two files. The data file is a sequence of
BlockSize byte blocks one after another, and the other file is a
locator file. The locator file's name is the name used by the FileName
property and the data file name has an extra '_' as first character.
Records in the data file occupy groups of successive blocks called a
group. Short records have just a few groups and long record have many
groups. The locator file keeps track of the groups and what they
contain.

Except for the first 4 longints the locator file is just a long list
of start/count pairs.

The first longints are:

  RecCount:  The number of records in the file. This number is
  incremented whenever a new record is inserted into the file and
  decremented whenever a record is deleted.

  BlockSiz:  The number of bytes in a block. Set when the file was
  created and is never changed.

  FreeCapacity: The number of possible entries in the FreeList discussed
  below.

  FreeCount: The actual number of entries in the Free List. Set when
  the file is created or when COMPRESS runs.

  MaxCount: The actual number of entries that have ever been on the
  FreeList since the file was created or compressed. COMPRESS uses this
  to suggest a new value for FreeCapacity.


Then there follows an array [1 .. FreeCapacity] of start/count pairs that
form the free list. This is a list of places in the file that used to
contain records but now those records have been deleted or moved. If the
pair (4507,17) occurs in the free area, it means that blocks 4507-4523
are available for storage. A record as long as 17*BlockSize bytes could
be stored there. When ever a new record is inserted or an old record must
be moved because it has grown too long for its current block this list
is consulted. If an existing free area is available it will be used
to house the record and otherwise the file will be extended at its end.
The pairs are kept is order of the start field so that abuting free areas
can be ccombined. The number of entries in the free list is kept in the
field FreeCount and can range from 0 to FreeCapacity. If there is no room
to store a new item on the Free List the item is discarded. The COMPRESS
utility reclaims these lost areas.

Finally there is the list of start/count pairs that determine where
the records are. For example, if (8305,9) is the pair with index 622
it means that record (622 - FreeCapacity-1) occupies blocks
8305-8313 at the present time. }




{============= These routines are useful later ================}

function TVFile.FindArea(BlocksWanted: longint): longint;
{Finds area in data file with room for BlocksWanted blocks and
returns the block number where the area starts. }

var
  i: word;

begin
  Result := FileSize(DataF);
  i := 0;
  while i < FreeCount do
    if FLP^[i].count = BlocksWanted
    then begin
      Result := FLP^[i].start;
      if i < FreeCount
      then move(FLP^[i+1],FLP^[i],8*(FreeCount-i));
      dec(FreeCount);
      exit;
      end
    else if FLP^[i].count > BlocksWanted
         then begin
           Result := FLP^[i].start;
           FLP^[i].start := FLP^[i].start + BlocksWanted;
           FLP^[i].count := FLP^[i].count - BlocksWanted;
           exit;
           end
         else inc(i);
  end;


procedure TVFile.AddToFreeList(GivenStart,GivenCount: longint);
{Adds a free space to the free list. }

var
  i: word;

begin
  if FreeCount = FreeCapacity
  then exit; {no room for another entry}
  if FreeCount = 0
  then begin
    FLP^[0].start := GivenStart;
    FLP^[0].count := GivenCount;
    inc(FreeCount);
    inc(MaxCount);
    exit;
    end;
  i := 0;
  while (i < FreeCount) and (FLP^[i].start < GivenStart) do inc(i);
  if i <= FreeCount
  then move(FLP^[i],FLP^[i+1],(FreeCount-i+1)*8);
  FLP^[i].start := GivenStart;
  FLP^[i].count := GivenCount;
  inc(FreeCount);
  if FreeCount > MaxCount
  then MaxCount := FreeCount;
  {Now collapse if possible}
  if i = 0
  then i := 1;
  while (FLP^[i-1].start+FLP^[i-1].count = FLP^[i].start)
  do begin
    FLP^[i-1].count := FLP^[i-1].count+FLP^[i].count;
    if i+1<=FreeCount
    then move(FLP^[i+1],FLP^[i],(FreeCount-i)*8);
    dec(FreeCount);
    end;
  end;




{============= Set property routines ==================}

procedure TVFile.SetVFileName(Value: string);

begin
  if FFileName <> Value
  then begin
    if Value = ''
    then if FAttached
         then DetachFile
         else
    else begin
      AssignFile(DataF,ExtractFileDir(Value)+
                    '_'+ExtractFileName(Value));
      {$I-} reset(DataF); {$I-}
      if IOResult <> 0
      then begin
        ShowMessage('The file '+ExtractFileDir(Value)+
           '_'+ExtractFileName(Value)+' does not exist.');
        FFileName := '';
        exit;
        end;
      CloseFile(DataF);
      AssignFile(LocF,Value);
      {$I-} reset(LocF,8); {$I+}
      if IOResult <> 0
      then begin
        ShowMessage('The file '+Value+' does not exist');
        FFileName := '';
        exit;
        end;
      BlockRead(LocF,RecCount,1);
      BlockSize := BlockSiz;
      CloseFile(LocF);
      end;;
    FFileName := Value;
    end;
  end;


procedure TVFile.SetBlockSize(Value: longint);

begin
  end;


{=============== TVFile methods =========================}

constructor TVFile.Create(AOwner: TComponent);

begin
  inherited Create(AOwner);
  FFileName := '';
  FAttached := FALSE;
  end;


destructor TVFile.Destroy;

begin
  DetachFile;
  inherited Destroy;
  end;



procedure TVFile.AttachFile;

begin
  if FAttached
  then begin
    ShowMessage('File already attached');
    exit;
    end;
  FAttached := FALSE; {Assume the worst}
  if FFileName = ''
  then begin
    ShowMessage('No file named yet.');
    exit;
    end;
  AssignFile(LocF,FFileName);
  reset(LocF,8);
  seek(LocF,0);
  BlockRead(LocF,RecCount,3);
  BlockSize := BlockSiz;
  GetMem(FLP,FreeCapacity*sizeof(LocItem));
  BlockRead(LocF,FLP^,FreeCapacity);
  AssignFile(DataF,ExtractFileDir(FFileName)+'_'+
        ExtractFileName(FFileName));
  reset(DataF,BlockSiz);
  FAttached := TRUE;
  end;


procedure TVFile.DetachFile;

begin
  if not FAttached
  then exit;
  if FFileName <> ''
  then begin
    seek(LocF,0);
    BlockWrite(LocF,RecCount,3);
    BlockWrite(LocF,FLP^,FreeCapacity);
    FreeMem(FLP,FreeCapacity*sizeof(LocItem));
    CloseFile(Locf);
    CloseFile(DataF);
    end;
  FFileName := '';
  FAttached := FALSE;
  end;


function TVFile.Exist(rid: longint): boolean;
{As a side effect this routine loads the locater item into GenLI.}

begin
  Result := TRUE; {Suppose OK}
  if (FreeCapacity+3+rid) >= filesize(LocF)
  then Result := FALSE
  else begin
    seek(LocF,FreeCapacity+3+rid);
    BlockRead(Locf,GenLI,1);
    if GenLI.count = 0
    then Result := FALSE;
    end;
  end;


function TVFile.GetBufSize(rid: longint): longint;

begin
  if not Exist(rid)
  then begin
    Result := 0;
    exit;
    end;
  Result := GenLI.count*BlockSiz;
  end;


function TVFile.Fetch(rid: longint; Buf: pointer): boolean;

begin
  Result := Exist(rid);
  if Exist(rid)
  then begin
    seek(DataF,GenLI.start);
    BlockRead(DataF,Buf^,GenLI.count);
    end;
  end;



function TVFile.Insert(rid: longint; Buf: pointer;
                       RL: longint): boolean;

var
  BlocksNeeded: longint;
  LI: LocItem;

begin
  Result := not Exist(rid);
  if Result
  then begin
    BlocksNeeded := (RL + BlockSiz - 1) div BlockSiz;
    LI.start := FindArea(BlocksNeeded);
    LI.count := BlocksNeeded;
    seek(DataF,LI.start);
    BlockWrite(DataF,Buf^,BlocksNeeded);
    inc(RecCount);
    {insert into locator file}
    seek(LocF,FreeCapacity+3+rid);
    BlockWrite(LocF,LI,1);
    end;
  end;



function TVFile.Restore(rid: longint; Buf: pointer;
                        RL: longint): boolean;

var
  BlocksNeeded: word;
  LI: LocItem;

begin
  Result := Exist(rid);
  if Result
  then begin
    {free the old area}
    AddToFreeList(GenLI.start,GenLI.count);
    BlocksNeeded := (RL + BlockSiz - 1) div BlockSiz;
    {now get a new one}
    LI.start := FindArea(BlocksNeeded);
    LI.count := BlocksNeeded;
    seek(DataF,LI.start);
    BlockWrite(DataF,Buf^,BlocksNeeded);
    seek(LocF,FreeCapacity+3+rid);
    BlockWrite(LocF,LI,1);
    end;
  end;


function TVFile.Delete(rid: longint): boolean;

var
  Dummy: LocItem;

begin
  Result := Exist(rid);
  if Result
  then begin
    {Free the old area}
    AddToFreeList(GenLI.start,GenLI.count);
    Dummy.start := -1; Dummy.count := 0;
    seek(LocF,FreeCapacity+3+rid);
    BlockWrite(Locf,Dummy,1);
    dec(RecCount);
    end;
  end;

end.

